"""
multiai - A Python library for text-based AI interactions with multi-provider support.
"""
import anthropic
import configparser
import enum
from google import genai
import json
import ollama
import openai
import os
import mistralai
import pypdf
import pyperclip
import requests
import sys
import trafilatura
from io import BytesIO
from importlib.metadata import distribution, PackageNotFoundError
from .printlong import print_long
import docx  # python-docx
from docx import Document

__all__ = [
    "Prompt",
    "Provider",
    "ColorCode",
]


class Prompt():
    """
    The Prompt main application.

    Usage:
        client = Prompt()
        answer = client.ask(prompt)
    """

    def __init__(self):
        # Values independent of system or user setting file
        self.role = 'user'
        # Anthropic requires max_tokens, so default value is given.
        # It can be overwritten by max_tokens.
        self.max_tokens_anthropic = 4096
        self.ai_providers = []
        # Load package data
        try:
            dist = distribution('multiai')
            self.version = dist.version
            md = dist.metadata  # email.message.Message
            self.description = (md.get('Summary') or '').strip()
            # Project-URL: Homepage, https://... or Home-page
            self.url = None
            for item in md.get_all('Project-URL') or []:
                label, _, link = item.partition(', ')
                if label.strip().lower() == 'homepage' and link:
                    self.url = link.strip()
                    break
            if not self.url:
                self.url = (md.get('Home-page') or '').strip() or None
        except PackageNotFoundError:
            self.version = None
            self.description = None
            self.url = None
        # Load user setting from config file in the order of
        # data/system.ini, ~/.multiai, .multai
        # It overwrites the system default values
        inifile = configparser.ConfigParser()
        here = os.path.abspath(os.path.dirname(__file__))
        inifile.read(os.path.join(here, 'data/system.ini'))
        conf_file = os.path.expanduser('~/.multiai')
        inifile.read(conf_file)
        inifile.read('.multiai')
        self.set_provider(inifile.get('model', 'ai_provider'))
        for provider in Provider:
            name = provider.name.lower()
            model = inifile.get('model', name, fallback=None)
            if model is None:
                print(f'multiai system error: {name} not found in [model].')
                sys.exit(1)
            setattr(self, 'model_' + name, model)
        self.temperature = inifile.getfloat('default', 'temperature')
        self.max_requests = inifile.getint('default', 'max_requests')
        self.blank_lines = inifile.getint('command', 'blank_lines')
        prompt_color = inifile.get('prompt', 'color')
        self.always_copy = inifile.getboolean('command', 'always_copy')
        self.copy = self.always_copy
        self.always_log = inifile.getboolean('command', 'always_copy')
        self.log = self.always_log
        try:
            self.prompt_color = ColorCode[prompt_color.upper()].value
        except Exception:
            print(f'Error in the settings file: color = {prompt_color}')
            available_colors = [name.lower()
                                for name in ColorCode.__members__.keys()]
            print(f'Available colors: {", ".join(available_colors)}')
            sys.exit(1)
        # No system default value is given from here.
        # Default values are given by fallback values.
        self.max_tokens = inifile.getint(
            'default', 'max_tokens', fallback=None)

        # From version 1.4.0: attachment character limit (per attachment)
        self.attach_char_limit = inifile.getint(
            'default', 'attach_char_limit', fallback=40000)

        for provider in Provider:
            env = os.getenv(provider.name + '_API_KEY')
            name = provider.name.lower()
            key = name + '_api_key'
            if env is None:
                ini = inifile.get('api_key', name, fallback=None)
                setattr(self, key, ini)
            else:
                setattr(self, key, env)

        # --- Azure TTS (not part of Provider) ---
        # Azure Text-to-Speech uses Azure Speech Services, which is a separate service
        # from Azure OpenAI. It requires its own dedicated API key and region.
        # The Azure OpenAI API key cannot be used for Speech (TTS/STT), so we load
        # the Speech-specific key explicitly and keep it independent from LLM
        # providers.

        # API key
        env = os.getenv('AZURE_TTS_API_KEY')
        if env is None:
            self.azure_tts_api_key = inifile.get(
                'api_key', 'azure_tts', fallback=None)
        else:
            self.azure_tts_api_key = env

        # region
        env = os.getenv('AZURE_TTS_REGION')
        if env is None:
            self.azure_tts_region = inifile.get(
                'azure_tts', 'region', fallback=None)
        else:
            self.azure_tts_region = env

        self.clear()

    def set_provider(self, provider):
        """
        Set AI provider.

        :param provider: str
            AI provider (case insensitive)
        """
        try:
            self.ai_provider = Provider[provider.upper()]
        except Exception:
            print(f'AI provider "{provider}" is not available.')
            sys.exit(1)

    def set_model(self, provider, model):
        """
        Set model.

        :param provider: str
            AI provider (case insensitive)
        :param model: str
            AI model
        """
        self.set_provider(provider)
        self.model = model
        setattr(self, 'model_' + provider.lower(), model)

    def clear(self):
        """
        Clear chat history.
        """
        self.openai_messages = []
        self.anthropic_messages = []
        self.google_messages = []
        self.perplexity_messages = []
        self.deepseek_messages = []
        self.mistral_messages = []
        self.xai_messages = []
        self.local_messages = []

    def ask(self, prompt, request=1, verbose=False):
        """
        Ask a question to AI.

        :param prompt: str
            Prompt to ask AI
        :param request: int
            Numbers of repetitive requests when the response is cut by token limit
        :param verbose: bool
            Show repeat process
        :return: str
            Answer from AI
        """
        self.message = [
            {
                "role": self.role,
                "content": prompt,
            }
        ]
        self.prompt = prompt
        if request == 1:
            self.prompt_continue = False
        else:
            self.prompt_continue = True
        # For example, call ask_openai() for openai
        func_name = 'ask_' + self.ai_provider.name.lower()
        try:
            func = getattr(self, func_name)
        except AttributeError:
            print(
                f'multiai system error: {func_name}() function is not defined.')
            sys.exit(1)
        func()
        # Error
        if self.error:
            return self.error_message
        # Finish successfully
        if self.finish_reason in ['stop', 'end_turn']:
            return self.response
        # Unexpected finish reason
        if self.finish_reason not in ['length', 'max_tokens']:
            self.response += f'\n\nFinish reason: {self.finish_reason}'
            return self.response
        # Response not finished. Continue the request.
        request += 1
        if request > self.max_requests:
            self.response += '\n\nFinished because of max_tokens and max_requests.'
            return self.response
        if verbose:
            print(
                f'{
                    self.color("Repeating...")} max_requests = {
                    self.max_requests}, requests = {request}\r',
                end='')
        response = self.response
        answer = self.ask('continue', request=request, verbose=verbose)
        if self.error:
            return answer
        return response + answer

    def ask_once(self, prompt):
        """
        Ask a single-turn question without polluting the chat history.

        This calls the same provider but restores all internal message lists
        after the request completes.

        :param prompt: str
            Prompt to ask AI
        :return: str
            Answer from AI
        """
        # Backup histories
        backups = {
            'openai_messages': list(self.openai_messages),
            'anthropic_messages': list(self.anthropic_messages),
            'google_messages': list(self.google_messages),
            'perplexity_messages': list(self.perplexity_messages),
            'deepseek_messages': list(self.deepseek_messages),
            'mistral_messages': list(self.mistral_messages),
            'xai_messages': list(self.xai_messages),
            'local_messages': list(self.local_messages),
        }
        # Ask
        answer = self.ask(prompt)
        # Restore histories regardless of error
        self.openai_messages = backups['openai_messages']
        self.anthropic_messages = backups['anthropic_messages']
        self.google_messages = backups['google_messages']
        self.perplexity_messages = backups['perplexity_messages']
        self.deepseek_messages = backups['deepseek_messages']
        self.mistral_messages = backups['mistral_messages']
        self.xai_messages = backups['xai_messages']
        self.local_messages = backups['local_messages']
        return answer

    def summarize_text(self, text, max_words_hint=600):
        """
        Summarize a long piece of text.

        Uses ask_once() to avoid altering conversation history.

        :param text: str
            Raw text to summarize
        :param max_words_hint: int
            A rough upper bound to guide the summary length
        :return: str
            Summarized text (best-effort)
        """
        prompt = (
            "Summarize the following content concisely. Preserve key facts, structure, and any code blocks. "
            f"Target up to roughly {max_words_hint} words. Do not include commentary about being an AI.\n\n"
            "Content begins below:\n\n"
            f"{text}"
        )
        return self.ask_once(prompt)

    def ask_print(self, prompt, prompt_summary=None):
        """
        Ask a question to AI and print, copy, log.

        :param prompt: str
            Prompt to ask AI
        :param prompt_summary: str
            Prompt shortened for logging
        """
        print(f'{self.color("Please wait ......")}\r', end='')
        if len(self.ai_providers) == 1:
            answer = self.ask(prompt, verbose=True)
            print(' ' * 50 + '\r', end='')
            if self.error:
                print(f'{self.color("Error message")}> {answer}')
                sys.exit(1)
            print(f'{self.color(self.model)}>')
            if self.log:
                if prompt_summary is not None:
                    prompt = prompt_summary
                try:
                    with open(self.log_file, mode='a') as f:
                        f.write(
                            f'### {self.role}:\n{prompt}\n### {self.model}:\n{answer}\n')
                except Exception as e:
                    print(e)
                    print('Check the setting of log_file.')
                    sys.exit(1)
        else:
            answer = ''
            if prompt_summary is None:
                prompt_log = prompt
            else:
                prompt_log = prompt_summary
            for provider in (self.ai_providers):
                self.ai_provider = provider
                single_answer = self.ask(prompt, verbose=True)
                model = getattr(self, 'model_' + provider.name.lower(), None)
                if self.error:
                    print(
                        f'{self.color("Error message from " + provider.name.lower())}> {single_answer}')
                    sys.exit(1)
                answer += f'### {model}:\n{single_answer}\n\n'
            answer = answer.strip()
            if self.log:
                with open(self.log_file, mode='a') as f:
                    f.write(
                        f'### {self.role}:\n{prompt_log}\n{answer}\n')
        print(' ' * 50 + '\r', end='')
        print_long(answer)
        if self.copy:
            pyperclip.copy(answer)

    def interactive(self, pre_prompt=''):
        """
        Interactive mode.

        :param pre_prompt: str
            Pre-prompt to append before prompt
        """
        prompt = ''
        blank = 0
        b = self.blank_lines
        if len(self.ai_providers) == 0:
            self.ai_providers = [self.ai_provider]
        if b > 0:
            print(
                f'\nInput {b} blank line{"s" if b > 1 else ""} to finish input.')
        while True:
            try:
                if prompt == '':
                    line = input(f'{self.color(self.role)}> ')
                else:
                    line = input()
            except EOFError:
                sys.exit()
            except KeyboardInterrupt:
                sys.exit()
            if line == '':
                if prompt == '':
                    print('Blank text entered. Enter "q" to quit.')
                    continue
                blank += 1
                if blank < self.blank_lines:
                    prompt += line + '\n'
                else:
                    self.ask_print(pre_prompt + prompt.strip())
                    prompt = ''
                    blank = 0
            elif prompt == '' and line in ['q', 'x', 'quit', 'exit']:
                sys.exit()
            else:
                if self.blank_lines == 0:
                    self.ask_print(pre_prompt + line)
                    prompt = ''
                else:
                    prompt += line + '\n'
                    blank = 0

    def color(self, text):
        """
        Return colored text with color defined at self.prompt_color.

        :param text: str
            Text
        :return: str
            Colored text (no color if not TTY)
        """
        if sys.stdout.isatty():
            return f'\033[{self.prompt_color}m{text}\033[0m'
        else:
            return text

    def retrieve_from_url(self, url, verbose=True):
        """
        Retrieve text from URL.

        When URL ends with ".pdf", PDF file is converted to text.

        :param url: str
            URL to retrieve data from
        :param verbose: bool
            Whether to print message
        :return: str
            Retrieved text
        """
        if verbose:
            print('Retrieving ...\r', end='')
        headers = {
            'User-Agent': self.user_agent if hasattr(self, 'user_agent') else None}
        try:
            response = requests.get(url, headers=headers)
        except Exception as e:
            if verbose:
                print(e)
            sys.exit(1)
        if response.status_code != 200:
            if verbose:
                print(f'{response.status_code} - {response.reason}')
            sys.exit(1)
        if verbose:
            print('Converting to text.\r', end='')
        if url.lower().endswith('.pdf'):
            with BytesIO(response.content) as pdf_file:
                reader = pypdf.PdfReader(pdf_file)
                text = ""
                for page in range(len(reader.pages)):
                    text += reader.pages[page].extract_text()
        else:
            text = trafilatura.extract(response.text)
            if text is None:
                if verbose:
                    print(f'{url} could not be retrieved.')
                sys.exit(1)
        return text

    def retrieve_from_file(self, source, filename=None, verbose=True):
        """
        Retrieve text from a file path or bytes.

        Supported extensions (special handling): txt, md, pdf, docx, html/htm, csv.
        For unknown extensions, if the content is valid UTF-8 text (no NUL bytes and UTF-8 decodable),
        it will be treated as plain text; otherwise an error is raised.

        :param source: str | bytes | file-like
            File path (str) or in-memory bytes (e.g., from upload).
        :param filename: str | None
            Original file name (used for extension detection and messages).
        :param verbose: bool
            Whether to print progress messages
        :return: str
            Extracted text
        """
        def _ext_from_name(name):
            return os.path.splitext(name)[1].lower()

        # Determine extension and load bytes
        if isinstance(source, (str, os.PathLike)):
            path = os.fspath(source)
            ext = _ext_from_name(path)
            if verbose:
                print('Reading file ...\r', end='')
            with open(path, 'rb') as f:
                data = f.read()
            name = os.path.basename(path)
        else:
            # in-memory data
            if hasattr(source, 'read'):
                data = source.read()
            else:
                data = source
            if not isinstance(data, (bytes, bytearray)):
                print('retrieve_from_file expects bytes/file-like for in-memory data.')
                sys.exit(1)
            if not filename:
                print('filename is required when passing in-memory data.')
                sys.exit(1)
            name = filename
            ext = _ext_from_name(filename)

        if verbose:
            print('Converting file to text.\r', end='')

        try:
            if ext in ['.txt', '.md', '.csv']:
                # Known simple text types: decode as UTF-8 (replace errors to
                # avoid crash)
                text = data.decode('utf-8', errors='replace')

            elif ext == '.pdf':
                with BytesIO(data) as pdf_file:
                    reader = pypdf.PdfReader(pdf_file)
                    text = ""
                    for page in range(len(reader.pages)):
                        text += reader.pages[page].extract_text()

            elif ext == '.docx':
                if Document is None:
                    print('python-docx is not installed. Please install "python-docx".')
                    sys.exit(1)
                with BytesIO(data) as stream:
                    doc = Document(stream)
                    text = "\n".join(p.text for p in doc.paragraphs)

            elif ext in ['.html', '.htm']:
                html = data.decode('utf-8', errors='replace')
                text = trafilatura.extract(html)
                if text is None:
                    print(f'{name} could not be converted from HTML.')
                    sys.exit(1)

            else:
                # Unknown extension: treat as text only if it is valid UTF-8
                # and not binary.
                if b'\x00' in data:
                    print(
                        f'Unsupported file type or binary content detected: {name}')
                    sys.exit(1)
                try:
                    # strict; will fail if not UTF-8
                    text = data.decode('utf-8')
                except UnicodeDecodeError:
                    print(
                        f'{name} is not a supported file format or not UTF-8 text.')
                    sys.exit(1)

        except Exception as e:
            print(f'Failed to parse {name}: {e}')
            sys.exit(1)

        return text

    # Implementations for each providers
    def ask_openai(self):
        """
        Ask a question to OpenAI.
        """
        if self.openai_api_key is None:
            self.error = True
            self.error_message = 'API key for OpenAI is not set.'
            return
        openai.api_key = self.openai_api_key
        if not self.prompt_continue:
            self.openai_messages += self.message
        try:
            if self.model[:5] == 'gpt-5' or self.model[0] == 'o':
                # gpt-5, o1 and o3 series do not support variable temperature and
                # max_tokens is written as max_completion_tokens
                self.completion = openai.chat.completions.create(
                    messages=self.openai_messages,
                    model=self.model_openai,
                    max_completion_tokens=self.max_tokens
                )
            else:
                self.completion = openai.chat.completions.create(
                    messages=self.openai_messages,
                    model=self.model_openai,
                    temperature=self.temperature,
                    max_tokens=self.max_tokens
                )
            self.error = False
            self.response = self.completion.choices[0].message.content.strip(
            )
            self.finish_reason = self.completion.choices[0].finish_reason
            self.openai_messages += [{"role": "assistant",
                                      "content": self.response}]
        except openai.APIError as e:
            self.error = True
            try:
                self.error_code = e.status_code
                self.error_dict = e.body
                self.error_type = f"Error {
                    self.error_code}: {
                    self.error_dict['code']}"
                self.error_message = f"{
                    self.error_type}\n{
                    self.error_dict['message']}"
            except Exception:
                self.error_message = e

    def ask_anthropic(self):
        """
        Ask a question to Anthropic.
        """
        if self.anthropic_api_key is None:
            self.error = True
            self.error_message = 'API key for Anthropic is not set.'
            return
        client = anthropic.Anthropic(api_key=self.anthropic_api_key)
        if not self.prompt_continue:
            self.anthropic_messages += self.message
        try:
            self.completion = client.messages.create(
                messages=self.anthropic_messages,
                model=self.model_anthropic,
                temperature=self.temperature,
                max_tokens=self.max_tokens if self.max_tokens else self.max_tokens_anthropic
            )
            self.error = False
            self.response = self.completion.content[0].text.strip()
            self.finish_reason = self.completion.stop_reason
            self.anthropic_messages += [{"role": "assistant",
                                         "content": self.response}]
        except Exception as e:
            self.error = True
            try:
                self.error_code = e.status_code
                self.error_dict = e.body['error']
                self.error_type = f"{
                    self.error_code}: {
                    self.error_dict['type']}"
                self.error_message = f"{
                    self.error_type}\n{
                    self.error_dict['message']}"
            except Exception:
                self.error_message = e

    def ask_google(self):
        """
        Ask a question to Google (google.genai).
        """
        # Suppress logging warnings of libraries
        os.environ["GRPC_VERBOSITY"] = "ERROR"
        os.environ["GLOG_minloglevel"] = "2"

        if self.google_api_key is None:
            self.error = True
            self.error_message = 'API key for Google is not set.'
            return

        # google.genai client
        client = genai.Client(api_key=self.google_api_key)

        if not hasattr(self, "google_messages"):
            self.google_messages = []

        if not self.prompt_continue:
            # {"role":"user","content":...}
            self.google_messages += self.message

        contents = []
        for m in self.google_messages:
            role = m.get("role", "user")
            if role == "assistant":
                role = "model"
            elif role == "system":
                role = "user"

            contents.append({
                "role": role,
                "parts": [{"text": m.get("content", "")}],
            })

        # generation config
        config = {}
        if self.temperature is not None:
            config["temperature"] = float(self.temperature)
        if self.max_tokens is not None:
            # google.genai は max_output_tokens
            config["max_output_tokens"] = int(self.max_tokens)

        try:
            resp = client.models.generate_content(
                model=self.model_google,
                contents=contents,
                config=config if config else None,
            )
            self.error = False
            text = getattr(resp, "text", None)
            if not text:
                text = ""
                try:
                    cand0 = resp.candidates[0]
                    parts = cand0.content.parts
                    text = "".join(getattr(p, "text", "") for p in parts)
                except Exception:
                    pass

            self.response = (text or "").replace('• ', '* ').strip()

            # finish_reason
            self.finish_reason = "stop"
            try:
                fr = resp.candidates[0].finish_reason
                name = getattr(fr, "name", None)
                if name:
                    self.finish_reason = str(name).lower()
                else:
                    s = str(fr).strip().lower()
                    if "." in s:
                        s = s.split(".")[-1]
                    self.finish_reason = s
            except Exception:
                pass

            self.google_messages += [{"role": "assistant",
                                      "content": self.response}]

        except Exception as e:
            self.error = True
            self.error_message = str(e)

    def ask_perplexity(self):
        """
        Ask a question to perplexity.
        """
        if self.perplexity_api_key is None:
            self.error = True
            self.error_message = 'API key for Perplexity is not set.'
            return
        base_url = 'https://api.perplexity.ai'
        client = openai.OpenAI(
            api_key=self.perplexity_api_key,
            base_url=base_url)
        self.perplexity_messages += self.message
        try:
            self.completion = client.chat.completions.create(
                messages=self.perplexity_messages,
                model=self.model_perplexity,
                temperature=self.temperature,
                max_tokens=self.max_tokens
            )
            self.error = False
            self.response = self.completion.choices[0].message.content.strip(
            )
            self.finish_reason = self.completion.choices[0].finish_reason
            self.perplexity_messages += [{"role": "assistant",
                                          "content": self.response}]
        except openai.APIError as e:
            self.error = True
            try:
                message = trafilatura.extract(e.message)
                self.error_message = message.splitlines()[0]
            except Exception:
                self.error_message = e

    def ask_deepseek(self):
        """
        Ask a question to DeepSeek.
        """
        if self.deepseek_api_key is None:
            self.error = True
            self.error_message = 'API key for DeepSeek is not set.'
            return
        base_url = 'https://api.deepseek.com'
        client = openai.OpenAI(
            api_key=self.deepseek_api_key,
            base_url=base_url)
        if not self.prompt_continue:
            self.deepseek_messages += self.message
        try:
            self.completion = client.chat.completions.create(
                messages=self.deepseek_messages,
                model=self.model_deepseek,
                temperature=self.temperature,
                max_tokens=self.max_tokens,
                stream=False
            )
            self.error = False
            self.response = self.completion.choices[0].message.content.strip(
            )
            self.finish_reason = self.completion.choices[0].finish_reason
            self.deepseek_messages += [{"role": "assistant",
                                        "content": self.response}]
        except json.JSONDecodeError:
            self.error = True
            self.error_message = 'Error: Invalid JSON response from DeepSeek API.'
        except openai.APIError as e:
            self.error = True
            try:
                self.error_code = e.status_code
                self.error_dict = e.body
                self.error_type = f"Error {
                    self.error_code}: {
                    self.error_dict['code']}"
                self.error_message = f"{
                    self.error_type}\n{
                    self.error_dict['message']}"
            except Exception:
                self.error_message = e

    def ask_mistral(self):
        """
        Ask a question to mistral.
        """
        if self.mistral_api_key is None:
            self.error = True
            self.error_message = 'API key for Mistral is not set.'
            return
        client = mistralai.Mistral(api_key=self.mistral_api_key)
        self.mistral_messages += self.message
        try:
            self.completion = client.chat.complete(
                messages=self.mistral_messages,
                model=self.model_mistral,
                temperature=self.temperature,
                max_tokens=self.max_tokens
            )
            self.error = False
            self.response = self.completion.choices[0].message.content.strip(
            )
            self.finish_reason = self.completion.choices[0].finish_reason
            self.mistral_messages += [{"role": "assistant",
                                       "content": self.response}]
        except mistralai.SDKError as e:
            self.error = True
            try:
                self.error_code = e.status_code
                self.error_dict = json.loads(e.body)
                self.error_message = f"Error {
                    self.error_code}: {
                    self.error_dict['message']}"
            except Exception:
                self.error_message = e

    def ask_xai(self):
        """
        Ask a question to xAI.
        """
        if self.xai_api_key is None:
            self.error = True
            self.error_message = 'API key for xAI is not set.'
            return
        base_url = 'https://api.x.ai/v1'
        client = openai.OpenAI(
            api_key=self.xai_api_key,
            base_url=base_url)
        self.xai_messages += self.message
        try:
            self.completion = client.chat.completions.create(
                messages=self.xai_messages,
                model=self.model_xai,
                temperature=self.temperature,
                max_tokens=self.max_tokens
            )
            self.error = False
            self.response = self.completion.choices[0].message.content.strip(
            )
            self.finish_reason = self.completion.choices[0].finish_reason
            self.xai_messages += [{"role": "assistant",
                                   "content": self.response}]
        except openai.APIError as e:
            self.error = True
            try:
                self.error_code = e.status_code
                self.error_type = f"Error {self.error_code}"
                self.error_message = f"{self.error_type}: {e.body}"
            except Exception:
                self.error_message = e

    def ask_local(self):
        """
        Ask a question to local language model.
        """
        if not self.prompt_continue:
            self.local_messages += self.message
        try:
            self.response = ollama.chat(
                messages=self.local_messages,
                model=self.model_local
            )
            self.error = False
            self.response = self.response.message.content.strip()
            self.finish_reason = 'stop'
            self.local_messages += [{"role": "assistant",
                                     "content": self.response}]
        except ConnectionError as e:
            self.error = True
            self.error_message = f'{e}\nInstall ollama and run "ollama serve".'
        except Exception as e:
            self.error = True
            try:
                self.error_code = e.status_code
                self.error_message = e.error
                if self.error_code == 404:
                    self.error_message += f'\nRun "ollama pull {
                        self.model}" and try again.'
            except Exception:
                self.error_message = e

    def list_models(self):
        """
        Retrieves and prints the list of available models for the currently selected AI provider.

        This method dynamically dispatches the call to a specific `list_models_<provider>`
        function based on the `self.ai_provider` attribute. If the specific function is
        not defined, it prints an error message and exits the program.
        """
        func_name = 'list_models_' + self.ai_provider.name.lower()
        try:
            func = getattr(self, func_name)
        except AttributeError:
            print(
                f'multiai system error: {func_name}() function is not defined.')
            sys.exit(1)
        models = func()
        for m in models:
            print(m)

    def _list_models_openai_compatible(self, api_key, base_url):
        """
        Helper method to fetch model lists from OpenAI-compatible APIs.

        Args:
            api_key (str): The API key for authentication.
            base_url (str): The base URL of the API endpoint.

        Returns:
            list[str]: A list of model IDs available at the endpoint. Returns an error message if an error occurs.
        """
        try:
            client = openai.OpenAI(api_key=api_key, base_url=base_url)
            models = []
            for m in client.models.list():
                models.append(m.id)
            return models
        except Exception as e:
            return [f"Error fetching models from {base_url}: {e}", ]

    def list_models_openai(self):
        """
        Retrieves the list of available models from the official OpenAI API.
        """
        return self._list_models_openai_compatible(
            api_key=self.openai_api_key,
            base_url="https://api.openai.com/v1"
        )

    def list_models_anthropic(self):
        """
        Returns a reference to Anthropic's model documentation.
        Note: Anthropic does not currently provide a standardized endpoint for listing models dynamically.
        """
        return [
            "See https://platform.claude.com/docs/en/about-claude/models/overview",
        ]

    def list_models_google(self):
        """
        Retrieves the list of available Gemini models via Google's OpenAI-compatible API endpoint.
        """
        base_url = "https://generativelanguage.googleapis.com/v1beta/openai/"

        return self._list_models_openai_compatible(
            api_key=self.google_api_key,
            base_url=base_url
        )

    def list_models_perplexity(self):
        """
        Returns a reference to Perplexity's model documentation.
        """
        return [
            "See https://docs.perplexity.ai/getting-started/models",
        ]

    def list_models_mistral(self):
        """
        Retrieves the list of available models from Mistral AI via their OpenAI-compatible API.
        """
        return self._list_models_openai_compatible(
            api_key=self.mistral_api_key,
            base_url="https://api.mistral.ai/v1"
        )

    def list_models_deepseek(self):
        """
        Retrieves the list of available models from DeepSeek via their OpenAI-compatible API.
        """
        return self._list_models_openai_compatible(
            api_key=self.deepseek_api_key,
            base_url="https://api.deepseek.com"
        )

    def list_models_xai(self):
        """
        Retrieves the list of available models from xAI (Grok) via their OpenAI-compatible API.
        """
        return self._list_models_openai_compatible(
            api_key=self.xai_api_key,
            base_url="https://api.x.ai/v1"
        )

    def list_models_local(self):
        """
        Returns a message indicating that listing local models is currently unavailable or not implemented.
        """
        return [
            "Unavailable to list models for local.",
        ]


class Provider(enum.Enum):
    """
    Provider is an Enum representing AI provider available at multiai.

    To add a provider definition,
    (1) Add the provider here. Note that the first letter should not
        overwrap other command-line options.
    (2) Define ask_provider() function in Prompt class.
    (3) Update clear() function in Prompt class.
    (4) Define default model at system.ini.
    """
    OPENAI = enum.auto()
    ANTHROPIC = enum.auto()
    GOOGLE = enum.auto()
    PERPLEXITY = enum.auto()
    MISTRAL = enum.auto()
    DEEPSEEK = enum.auto()
    XAI = enum.auto()
    LOCAL = enum.auto()


class ColorCode(enum.Enum):
    """
    ColorCode is an Enum representing ANSI color codes.

    Each member of this Enum corresponds to a specific color used in terminal output.
    """
    BLACK = 30
    RED = 31
    GREEN = 32
    YELLOW = 33
    BLUE = 34
    MAGENTA = 35
    CYAN = 36
    WHITE = 37
    BACK_BLACK = 40
    BACK_RED = 41
    BACK_GREEN = 42
    BACK_YELLOW = 43
    BACK_BLUE = 44
    BACK_MAGENTA = 45
    BACK_CYAN = 46
    BACK_WHITE = 47



from multiai.printlong import calculate_display_width, wrap_text

def test_calculate_display_width():
    # ASCII characters have width 1
    assert calculate_display_width("abc") == 3
    # Wide East Asian characters have width 2
    assert calculate_display_width("あいう") == 6
    # Mixed characters
    assert calculate_display_width("aあb") == 4

def test_wrap_text():
    text = "あいうえお"
    # Wrap at width 4 -> "あい", "うえ", "お" (Since "あ" is width 2)
    wrapped = wrap_text(text, 4)
    assert wrapped == ["あい", "うえ", "お"]
    
    text2 = "abcde"
    # Wrap at width 3 -> "abc", "de"
    wrapped2 = wrap_text(text2, 3)
    assert wrapped2 == ["abc", "de"]



"""
print_long - print long text with pypager
"""
import shutil
import unicodedata
import pypager


def print_long(text):
    """
    Print long text with pypager.

    When text fits in 1 page in a terminal, it uses print,
    otherwise it uses pypager.

    :param text: str
        text to display
    """
    default_terminal_size = (80, 20)
    terminal_size = shutil.get_terminal_size(default_terminal_size)
    lines_per_page = terminal_size.lines - 1
    terminal_width = terminal_size.columns
    wrapped_lines = []
    for line in text.split('\n'):
        wrapped_lines.extend(wrap_text(line, terminal_width))
    total_lines = len(wrapped_lines)
    wrapped_text = '\n'.join(wrapped_lines)
    if total_lines <= lines_per_page:
        print(text)
    else:
        p = pypager.pager.Pager()
        p.add_source(pypager.source.StringSource(wrapped_text))
        p.run()


def calculate_display_width(text):
    """
    Calculate display width of a line.

    :param text: str
        text to count
    """
    width = 0
    for char in text:
        if unicodedata.east_asian_width(char) in 'WF':
            width += 2
        else:
            width += 1
    return width


def wrap_text(text, width):
    """
    Wrap text of a line.

    :param text: str
        text to wrap
    :param width: int
        display width
    """
    lines = []
    current_line = ""
    current_width = 0
    for line in text.split('\n'):
        if not line:
            lines.append("")
            continue
        for char in line:
            char_width = 2 if unicodedata.east_asian_width(char) in 'WF' else 1
            if current_width + char_width > width:
                lines.append(current_line)
                current_line = char
                current_width = char_width
            else:
                current_line += char
                current_width += char_width
        if current_line:
            lines.append(current_line)
            current_line = ""
            current_width = 0
    if current_line:
        lines.append(current_line)
    return lines



"""
Entry point for multiai CLI with file attachment support.

Usage examples:
  ai "Explain this" -f doc1.pdf notes.md --attach-limit 50000
  ai -u https://example.com/article.html -f table.csv
  ai -e -f report.docx  # English helper + attachments
"""
import argparse
import configparser
import os
import readline
import subprocess
import sys
import webbrowser
from datetime import datetime
from .multiai import Prompt, Provider

__all__ = [
    "entry",
]


def entry():
    """
    Entry point of multiai CLI (invoked by `ai` command).
    """
    client = Prompt()
    # Load user setting from config file in the order of
    # data/system.ini, ~/.multiai, .multiai
    inifile = configparser.ConfigParser()
    here = os.path.abspath(os.path.dirname(__file__))
    inifile.read(os.path.join(here, 'data/system.ini'))
    inifile.read(os.path.expanduser('~/.multiai'))
    inifile.read('.multiai')
    # Start reading [command] section of the config file
    # log_file: file name of the log file.
    log_file = inifile.get('command', 'log_file')
    log_file = os.path.expanduser(log_file)
    log_file = log_file.replace('DATE', datetime.today().strftime('%Y%m%d'))
    client.log_file = log_file
    # user_agent: user agent when retrieving web data
    client.user_agent = inifile.get('command', 'user_agent', fallback=None)
    # [prompt] section
    prompt_english = inifile.get('prompt', 'english')
    prompt_url = inifile.get('prompt', 'url')

    # Load command-line arguments
    parser = argparse.ArgumentParser(
        description=f'multiai {client.version} - {client.description} {client.url}')
    parser.add_argument('prompt', nargs='*',
                        help='prompt for AI')
    for provider in Provider:
        name = provider.name.lower()
        help_msg = 'use ' + name
        if client.ai_provider == provider:
            help_msg += ' (Default)'
        parser.add_argument(
            # -i for mistral, -a for anthropic, etc.
            '-' + name.replace('m', '')[0],
            '--' + name,
            action='store_true',
            help=help_msg)
    parser.add_argument('-m', '--model',
                        help='set model')
    parser.add_argument('--list',
                        action='store_true', help='list models')
    parser.add_argument('-t', '--temperature',
                        help=f'set temperature. 0 is deterministic. Default is {client.temperature}.')
    parser.add_argument('-e', '--english',
                        action='store_true', help='correct if English, translate into English otherwise')
    parser.add_argument('-u', '--url',
                        help='retrieve text from the URL')

    # From version 1.4.0: file attachment option (replaces old -f/--factual)
    parser.add_argument('-f', '--file', nargs='+', action='append',
                        help='attach one or more files (txt, md, pdf, docx, html/htm, csv)')
    parser.add_argument('--attach-limit', type=int, default=client.attach_char_limit,
                        help=f'character limit per attachment before auto-summarization (default {client.attach_char_limit})')

    if not client.always_copy:
        parser.add_argument('-c', '--copy',
                            action='store_true', help='copy the latest answer')
    if not client.always_log:
        parser.add_argument('-s', '--save',
                            action='store_true', help=f'save log as {log_file}')
    args = parser.parse_args()

    # Set ai_provider, ai_providers and model
    client.ai_providers = []
    for provider in Provider:
        if getattr(args, provider.name.lower()):
            client.ai_provider = provider
            client.ai_providers.append(provider)
    if len(client.ai_providers) == 0:
        client.ai_providers = [client.ai_provider]
    if len(client.ai_providers) == 1:
        default_model = 'model_' + client.ai_provider.name.lower()
        if args.model:
            setattr(client, default_model, args.model)
        client.model = getattr(client, default_model, None)

    # List models
    if args.list:
        client.list_models()
        sys.exit()

    # Get prompt
    prompt = ' '.join(args.prompt).strip()

    # -t option
    if args.temperature:
        try:
            client.temperature = float(args.temperature)
        except ValueError:
            print("Invalid 'temperature': should be a number.")
            sys.exit(1)
        if client.temperature < 0:
            print("Invalid 'temperature': should be >=0.")
            sys.exit(1)

    # -c option
    if client.always_copy:
        args.copy = True
    client.copy = args.copy

    # -s option
    if client.always_log:
        args.save = True
    client.log = args.save
    if args.save:
        if not os.path.exists(log_file):
            try:
                with open(log_file, 'w') as file:
                    file.write("# AI chat log\n\n")
            except Exception as e:
                print(e)
                print('Check the setting of log_file.')
                sys.exit(1)

    # -e option
    pre_prompt = ''
    if args.english:
        pre_prompt = prompt_english + '\n\n'

    # Track if the user did not provide a prompt originally
    original_prompt_empty = (len(' '.join(args.prompt).strip()) == 0)

    # Build attachment text if any
    attached_texts = []
    attached_names = []
    if args.file:
        # Flatten nested lists
        all_files = [f for group in args.file for f in group]
        for fpath in all_files:
            if not os.path.exists(fpath):
                print(f'File not found: {fpath}')
                sys.exit(1)
            text = client.retrieve_from_file(fpath, verbose=True)
            n = len(text)
            if n > args.attach_limit:
                print(
                    f'Attachment "{os.path.basename(fpath)}" has {n} characters and exceeds the limit {args.attach_limit}. Summarizing the attachment.')
                text = client.summarize_text(text)
            attached_texts.append((os.path.basename(fpath), text))
            attached_names.append(os.path.basename(fpath))

    attachments_section = ''
    if attached_texts:
        blocks = []
        for name, text in attached_texts:
            blocks.append(f'=== Attachment: {name} ===\n{text}')
        attachments_section = '\n\n'.join(blocks)

    # One-shot send when prompt is empty and attachments are present
    did_one_shot_with_attachments = False
    if original_prompt_empty and attachments_section:
        base_prompt = prompt_url
        print(
            f'{client.color(client.role)}> {base_prompt}\n\nAttachments: {", ".join(attached_names)}')
        prompt_summary = f'{prompt_url}\n\nAttachments: {
            ", ".join(attached_names)}'
        one_shot_prompt = base_prompt + '\n' + attachments_section
        client.ask_print(one_shot_prompt, prompt_summary=prompt_summary)
        did_one_shot_with_attachments = True

    # -u option
    if args.url:
        text = client.retrieve_from_url(args.url)
        if prompt == '':
            prompt = prompt_url
        print(f'{client.color(client.role)}> {prompt}\n\nText of {args.url}')
        prompt_summary = f'{prompt_url}\n\nText of {args.url}'
        prompt += '\n' + text
        client.ask_print(prompt, prompt_summary=prompt_summary)

    # Finished loading arguments and run
    # Enter interactive mode if:
    # - no prompt was provided originally, or
    # - URL was used (keeps old behavior), or
    # - we already did a one-shot with attachments above.
    if original_prompt_empty or args.url or did_one_shot_with_attachments:
        client.interactive(pre_prompt=pre_prompt)
    else:
        # Prompt provided; attach attachments immediately and send once
        if attachments_section:
            # For logging, keep a concise summary line about attachments
            prompt_summary = f'{prompt}\n\nAttachments: {
                ", ".join(attached_names)}'
            prompt = pre_prompt + prompt + '\n\n' + attachments_section
            client.ask_print(prompt, prompt_summary=prompt_summary)
        else:
            client.ask_print(pre_prompt + prompt)



import pytest
from unittest.mock import MagicMock, patch, mock_open
from multiai.multiai import Prompt, Provider

class TestPromptCore:
    
    def test_initialization(self, mock_config, mock_environment):
        """Test if the Prompt class initializes correctly with mocked config."""
        client = Prompt()
        assert client.ai_provider == Provider.OPENAI
        assert client.model_openai == 'gpt-4o'
        assert client.openai_api_key == 'sk-dummy-openai'
        assert client.temperature == 0.7

    def test_set_provider(self, mock_config, mock_environment):
        """Test switching AI providers."""
        client = Prompt()
        client.set_provider('ANTHROPIC')
        assert client.ai_provider == Provider.ANTHROPIC

        # Test invalid provider calls sys.exit(1)
        with pytest.raises(SystemExit):
            client.set_provider('INVALID_PROVIDER')

    @patch('multiai.multiai.openai.chat.completions.create')
    def test_ask_openai(self, mock_create, mock_config, mock_environment):
        """Test the logic for asking OpenAI (mocking the API response)."""
        # Mock the OpenAI response object structure
        mock_response = MagicMock()
        mock_response.choices[0].message.content = "Hello, world!"
        mock_response.choices[0].finish_reason = "stop"
        mock_create.return_value = mock_response

        client = Prompt()
        client.set_provider('OPENAI')
        
        # FIX: The Prompt class requires .model to be set before calling ask_openai.
        client.model = client.model_openai

        response = client.ask("Hi")

        assert response == "Hello, world!"
        assert len(client.openai_messages) == 2  # user message + assistant response
        assert client.openai_messages[0]['content'] == "Hi"

    def test_retrieve_from_file_text(self, mock_config, tmp_path):
        """Test retrieving text from a standard text file."""
        client = Prompt()
        
        # Create a temporary text file
        p = tmp_path / "hello.txt"
        p.write_text("Hello file content", encoding="utf-8")
        
        text = client.retrieve_from_file(str(p), verbose=False)
        assert text == "Hello file content"

    # Changed from PyPDF2.PdfReader to pypdf.PdfReader
    @patch('multiai.multiai.pypdf.PdfReader')
    def test_retrieve_from_file_pdf(self, mock_pdf_reader, mock_config):
        """Test PDF extraction logic (mocking pypdf)."""
        client = Prompt()
        
        # Mock PDF reader behavior
        mock_page = MagicMock()
        mock_page.extract_text.return_value = "PDF Page Content"
        mock_pdf_instance = mock_pdf_reader.return_value
        mock_pdf_instance.pages = [mock_page]

        # Use mock_open to bypass actual file reading, passing bytes for PDF check
        with patch('builtins.open', mock_open(read_data=b'%PDF-1.4...')) as m:
            text = client.retrieve_from_file("dummy.pdf", verbose=False)
            assert "PDF Page Content" in text

    @patch('multiai.multiai.Prompt.ask_once')
    def test_summarize_text(self, mock_ask_once, mock_config, mock_environment):
        """Test the summarization helper."""
        client = Prompt()
        mock_ask_once.return_value = "Summarized content"
        
        long_text = "A" * 50000
        result = client.summarize_text(long_text)
        
        assert result == "Summarized content"
        mock_ask_once.assert_called_once()



import pytest
from unittest.mock import patch, MagicMock
import sys
from multiai.entry import entry
from multiai.multiai import Provider

class TestEntryCLI:

    @patch('multiai.entry.Prompt')
    @patch('multiai.entry.sys.argv', ['ai', 'hello world'])
    def test_entry_simple_prompt(self, MockPrompt, mock_config):
        """Test simple CLI usage: ai "hello world" """
        mock_client = MockPrompt.return_value
        # Mock default client attributes
        mock_client.ai_provider = Provider.OPENAI
        mock_client.ai_providers = []
        mock_client.always_copy = False
        mock_client.always_log = False
        mock_client.role = 'user'
        mock_client.color.return_value = '' 

        entry()

        # Ensure ask_print was called with the prompt
        mock_client.ask_print.assert_called()
        args, _ = mock_client.ask_print.call_args
        assert "hello world" in args[0]

    @patch('multiai.entry.Prompt')
    @patch('multiai.entry.sys.argv', ['ai', '-o', 'hello'])
    def test_entry_provider_flag(self, MockPrompt, mock_config):
        """Test provider flags: ai -o (OpenAI)"""
        mock_client = MockPrompt.return_value
        mock_client.ai_provider = Provider.OPENAI # default
        mock_client.ai_providers = []
        mock_client.always_copy = False
        mock_client.always_log = False
        
        entry()
        
        # Check if the CLI logic updated the ai_providers list
        assert Provider.OPENAI in mock_client.ai_providers

    @patch('multiai.entry.Prompt')
    @patch('multiai.entry.os.path.exists', return_value=True)
    @patch('multiai.entry.sys.argv', ['ai', 'summarize', '-f', 'doc.txt'])
    def test_entry_file_attachment(self, mock_exists, MockPrompt, mock_config):
        """Test file attachment: ai summarize -f doc.txt"""
        mock_client = MockPrompt.return_value
        mock_client.ai_provider = Provider.OPENAI
        mock_client.ai_providers = []
        mock_client.always_copy = False
        mock_client.always_log = False
        mock_client.attach_char_limit = 50000
        
        # Mock return value of file retrieval
        mock_client.retrieve_from_file.return_value = "File Content Here"

        entry()

        # Ensure retrieval was called
        mock_client.retrieve_from_file.assert_called_with('doc.txt', verbose=True)
        
        # Ensure the prompt passed to ask_print contains the attachment
        args, _ = mock_client.ask_print.call_args
        prompt_sent = args[0]
        assert "=== Attachment: doc.txt ===" in prompt_sent
        assert "File Content Here" in prompt_sent

    @patch('multiai.entry.Prompt')
    @patch('multiai.entry.sys.argv', ['ai']) # No arguments
    def test_entry_interactive_mode(self, MockPrompt, mock_config):
        """Test interactive mode when no prompt is provided."""
        mock_client = MockPrompt.return_value
        mock_client.ai_provider = Provider.OPENAI
        mock_client.ai_providers = []
        mock_client.always_copy = False
        mock_client.always_log = False

        entry()

        # interactive() method should be called
        mock_client.interactive.assert_called_once()
